# !/usr/bin/env Python3
# -*- coding:utf-8 -*-
# Team : HiLab
# Author：Yu Tao
# Date ：2022/6/22 9:54
# Tool ：PyCharm
# Email : yuao_11260103@163.com
# Project : msDAWG
# ------ ------
#  DualAE to construct feature
# ------ ------

import torch
import random
import numpy as np
from scripts.model import setup_seed, unsupv_AE
import torch.nn.functional as F
from torch.utils.data import DataLoader
from param_class import Dataset_Param
from load_data import split_train_test
from torch import nn, optim
import matplotlib.pyplot as plt
from matplotlib.pyplot import subplots
from datetime import datetime
# ------ dataset param ------
# HIP = Dataset_Param(0, 20, 1e-4, 100)
# ICC = Dataset_Param(1, 20, 1e-4, 50)
# SIMS = Dataset_Param(2, 20, 1e-4, 50)
# OvarianL162 = Dataset_Param(3, 20, 1e-4, 50)
# PanIN = Dataset_Param(4, 10, 1e-4, 100)
# OvarianL100 = Dataset_Param(5, 30, 1e-4, 100)
# JNCIL = Dataset_Param(6, 10, 1e-4, 100)
# OvarianH121 = Dataset_Param(7, 20, 1e-4, 50)
# ------ ------
dt = datetime.today()
file_name = str(dt.day) + '-' + str(dt.hour) + '-' + str(dt.minute)

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
# Assuming that we are on a CUDA machine, this should print a CUDA device:
print(device)
# net.to(device)

seed = 19
setup_seed(seed)
torch.manual_seed(seed)
torch.cuda.manual_seed_all(seed)
np.random.seed(seed)
random.seed(seed)
torch.backends.cudnn.deterministic = True


class MyDataSet(torch.utils.data.Dataset):
    def __init__(self, data, label):
        self.data = torch.from_numpy(data).float()
        self.label = torch.from_numpy(label).float()  # torch.from_numpy(label)
        self.length = data.shape[0]

    def __getitem__(self, index):
        return self.data[index], self.label[index]

    def __len__(self):
        return self.length


def train_module(dataloader, model, optimizer):
    num_batches = len(dataloader)
    train_loss = 0
    for i, data in enumerate(dataloader, 0):
        input, label = data
        input, label = input.to(device), label.to(device)
        _, _, _, _, output = model(input)

        loss = F.mse_loss(output, input)
        train_loss += loss.item()

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    return train_loss / num_batches


def test_module(dataloader, model):
    num_batches = len(dataloader)
    test_loss = 0
    with torch.no_grad():
        for i, data in enumerate(dataloader, 0):
            input, label = data
            input, label = input.to(device), label.to(device)
            _, _, _, _, output = model(input)
            loss = F.mse_loss(output, input)
            test_loss += loss.item()
    return test_loss / num_batches


def AE_feature_construct(X, y, Dataset_Param):
    print('---start AE_feature_construct---')

    # ========param===========
    data_index = Dataset_Param.index
    data_name = Dataset_Param.name
    param_batch_size = Dataset_Param.batch_size
    param_lr = Dataset_Param.lr
    param_epoch = Dataset_Param.epoch

    # ========================

    X_train, X_test, y_train, y_test = split_train_test(X, y)
    data_dict = {'X_train': X_train, 'X_test': X_test, 'y_train': y_train, 'y_test': y_test}

    print("data loading...")
    train_data_set = MyDataSet(data_dict['X_train'], data_dict['y_train'])
    train_dataloader = DataLoader(train_data_set, batch_size=param_batch_size, shuffle=True, drop_last=True)
    test_data_set = MyDataSet(data_dict['X_test'], data_dict['y_test'])
    test_dataloader = DataLoader(test_data_set, batch_size=param_batch_size, shuffle=True, drop_last=True)

    print('model constructing....')
    net = unsupv_AE(data_dict['X_train'].shape[1])
    net.to(device)
    optimizer = optim.Adam(net.parameters(), lr=param_lr)  # lr 0.001太大  1e-3  5e-3 1e-4  5e-4 1e-5 5e-5
    ls = []
    lss = []
    for i in range(param_epoch):  # epoch
        print(i)
        net.train()
        ls.append(train_module(train_dataloader, net, optimizer))
        net.eval()
        lss.append(test_module(test_dataloader, net))
    print("OK")

    # ---观察损失曲线---
    print('Start drawing....')
    fig, axes = subplots(1, 1)
    axes.plot(ls, label='train-loss')
    axes.plot(lss, label='test-loss')
    axes.set_xlabel('epoch', fontsize=13)
    plt.title(data_name + '-trainset-AE_feature_construct')
    plt.legend(fontsize=12)
    plt.savefig('fig/' + data_name + 'Sps' + file_name + '.png')

    torch.save(net, 'model/' + data_name + '_AE_model.pkl')
    return net
