# !/usr/bin/env Python3
# -*- coding:utf-8 -*-
# Team : HiLab
# Author：Yu Tao
# Date ：2022/6/20 20:18
# Tool ：PyCharm
# Email : yuao_11260103@163.com
# Project : msDAWG
import numpy as np
from sklearn.model_selection import StratifiedKFold
from sklearn.svm import LinearSVC
from sklearn.feature_selection import RFE
import scripts.fs_tools as fs
from sklearn.feature_selection import VarianceThreshold, SelectFromModel
import shap
import xgboost


def t_test(data_dict):
    """
    t-test feature rank

    :param data_dict
    :return:feature_index -> arr
    """
    print('-- start t_test_FS --')
    feature_index = fs.t_ranking(data_dict['X_train'], data_dict['y_train'])
    print('feature_index:', feature_index)
    print('feature_num:', len(feature_index))
    return feature_index


def w_test(data_dict):
    """
    w-test feature rank

    :param data_dict
    :return:feature_index -> arr
    """
    print('-- start w_test_FS --')
    feature_index = fs.w_ranking(data_dict['X_train'], data_dict['y_train'])
    print('feature_index:', feature_index)
    print('feature_num:', len(feature_index))
    return feature_index


def Variance_FS(data_dict):
    """
    filter：Variance feature select

    :param data_dict
    :return: feature_index -> arr
    """
    print('-- start Variance_FS --')
    selector = VarianceThreshold()
    selector.fit(data_dict['X_train'])
    feature_index = np.argsort(-selector.variances_)
    print('featrue_index:', feature_index)
    print('feature_num:', len(feature_index))
    return feature_index


def shap_select(data_dict, X):
    """
    use shap to feature select,use XGboost train model

    :param data_dict
    :param X:train_X+test_X
    :return:shap_ranked_index -> arr
    """
    print('-- start shap_FS --')
    model = xgboost.XGBClassifier(n_estimators=500, subsample=1)
    model.fit(data_dict['X_train'], data_dict['y_train'])
    shap_explainer = shap.TreeExplainer(model)
    shap_vals = shap_explainer.shap_values(X)
    shap_ranked_index = np.argsort(abs(shap_vals).mean(0))[::-1]
    # data_dict_selected = restructuring_dict(data_dict,shap_ranked_index)
    print('featrue_index:', shap_ranked_index)
    print('feature_num:', len(shap_ranked_index))
    return shap_ranked_index


def trivote(data, label, k=70):
    """
    trivote

    :param data: (sanmple,features)
    :param label:label
    :param k:default:70
    :return:selected_feature_matrix numpy
    """
    # step1 LinearSVC
    model = LinearSVC(dual=False, random_state=1)
    model.fit(data, label)
    sfm = SelectFromModel(estimator=model, threshold=-np.inf, max_features=int(k * 5), prefit=True)
    matrix = sfm.transform(data)
    # step2 SVM-RFE
    selector = RFE(estimator=LinearSVC(), n_features_to_select=k * 2)
    matrix = selector.fit_transform(matrix, label)
    # step3 LinearSVC
    max_fea = k  # LinearSVC
    t = 0  # loop
    matrix_temp = matrix  # init
    while t < 20:
        t += 1
        # select feature
        lsvc = LinearSVC(dual=False).fit(matrix, label)
        model = SelectFromModel(estimator=lsvc, threshold=-np.inf, max_features=max_fea, prefit=True)
        max_fea += 1
        matrix_temp = model.transform(matrix)
        if matrix_temp.shape[1] == 0:
            continue
        kf = StratifiedKFold(n_splits=10, random_state=1, shuffle=True)
        acc_list = []
        for train, val in kf.split(matrix_temp, label):
            train_data, train_label = np.array(matrix_temp)[train], np.array(label)[train]
            val_data, val_label = np.array(matrix_temp)[val], np.array(label)[val]
            clf = LinearSVC(dual=False)
            clf.fit(train_data, train_label)
            predicts = clf.predict(val_data)
            error = 0
            for i in range(len(predicts)):
                if predicts[i] != val_label[i]:
                    error += 1
            acc = round((len(predicts) - error) / len(predicts), 6)
            acc_list.append(acc)
        mean_acc = np.array(acc_list).mean()
        if mean_acc >= 1:
            break
    return matrix_temp


def SFM(model, data_dict, feature_num):
    """
    sklearn select from model

    :param model: estimator : feature_importances_ or coef_
    :param data_dict:dataset
    :param feature_num:max_features
    :return:selected_features_matrix
    """
    model.fit(data_dict['X_train'], data_dict['y_train'])
    sfm = SelectFromModel(estimator=model, threshold=-np.inf, max_features=feature_num, prefit=True)
    selected_features_matrix = sfm.transform(data_dict['X_train'])
    return selected_features_matrix

