# !/usr/bin/env Python3
# -*- coding:utf-8 -*-
# Team : HiLab
# Author：Yu Tao
# Date ：2022/6/20 20:16
# Tool ：PyCharm
# Email : yuao_11260103@163.com
# Project : msDAWG
import torch
from torch import nn
import torch.nn.functional as F
import numpy as np
import random

from torch.nn.modules.activation import Sigmoid

# ------ ------
#  construction of net
# ------ ------

def setup_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True


# ------ ------
# WGAN + AE +attention (Generator :dual-AE)
# ------ ------
class wGenerator_attention_dualAE(nn.Module):

    def __init__(self, N1):
        super(wGenerator_attention_dualAE, self).__init__()
        self.N1 = N1
        self.fc1 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc2 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc3 = nn.Linear(int(np.ceil(N1 / 25)), 2)
        self.fc4 = nn.Linear(2, int(np.ceil(N1 / 25)))
        self.fc5 = nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5)))
        self.fc6 = nn.Linear(int(np.ceil(N1 / 5)), N1)
        self.fc7 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc8 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc9 = nn.Linear(int(np.ceil(N1 / 25)), 2)

    def forward(self, x):
        out2 = self.fc1(x)
        out3 = self.fc2(out2)
        out_middle = self.fc3(out3)
        out5 = self.fc4(out_middle)
        out6 = self.fc5(out5)
        output = self.fc6(out6)
        out7 = self.fc7(output)
        out8 = self.fc8(out7)

        out_label = self.fc9(out8)
        return x, out2, out3, out_middle, out5, out6, output, out7, out8, out_label


class wDiscriminator_attention_dualAE(nn.Module):
    def __init__(self, N1):
        super(wDiscriminator_attention_dualAE, self).__init__()
        self.N1 = N1
        self.softmax = nn.Softmax(dim=1)
        self.sigmoid = nn.Sigmoid()
        self.fc1 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc2 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc3 = nn.Linear(int(np.ceil(N1 / 25)), 1)

    def attention(self, a, b, input):
        middle_result = self.softmax(torch.mm(a, b.T)) / np.sqrt(a.size()[1])
        result = torch.mm(middle_result, input)
        return result

    def forward(self, x):
        out1 = self.fc1(x)
        out1 = self.attention(out1, out1, out1)
        out2 = self.fc2(out1)
        output = self.fc3(out2)
        return output


# ------ ------
# WGAN + AE  (Generator :dual-AE)
# ------ ------
class wGenerator_dualAE(nn.Module):

    def __init__(self, N1):
        super(wGenerator_dualAE, self).__init__()
        self.N1 = N1
        self.fc1 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc2 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc3 = nn.Linear(int(np.ceil(N1 / 25)), 2)
        self.fc4 = nn.Linear(2, int(np.ceil(N1 / 25)))
        self.fc5 = nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5)))
        self.fc6 = nn.Linear(int(np.ceil(N1 / 5)), N1)
        self.fc7 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc8 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc9 = nn.Linear(int(np.ceil(N1 / 25)), 2)

    def forward(self, x):
        out2 = self.fc1(x)
        out3 = self.fc2(out2)
        out_middle = self.fc3(out3)
        out5 = self.fc4(out_middle)
        out6 = self.fc5(out5)
        output = self.fc6(out6)
        out7 = self.fc7(output)
        out8 = self.fc8(out7)

        out_label = self.fc9(out8)
        return x, out2, out3, out_middle, out5, out6, output, out7, out8, out_label


class wDiscriminator_dualAE(nn.Module):
    def __init__(self, N1):
        super(wDiscriminator_dualAE, self).__init__()
        self.N1 = N1
        self.fc1 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc2 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc3 = nn.Linear(int(np.ceil(N1 / 25)), 1)

    def forward(self, x):
        out1 = self.fc1(x)
        out2 = self.fc2(out1)
        output = self.fc3(out2)
        return output


# ------ ------
# GAN + AE + attention   (Generator :dual-AE)
# ------ ------
class Generator_attention_dualAE(nn.Module):

    def __init__(self, N1):
        super(Generator_attention_dualAE, self).__init__()
        self.N1 = N1

        self.fc1 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc2 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc3 = nn.Linear(int(np.ceil(N1 / 25)), 2)
        self.fc4 = nn.Linear(2, int(np.ceil(N1 / 25)))
        self.fc5 = nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5)))
        self.fc6 = nn.Linear(int(np.ceil(N1 / 5)), N1)
        self.fc7 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc8 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc9 = nn.Linear(int(np.ceil(N1 / 25)), 2)

    def forward(self, x):
        out2 = self.fc1(x)
        out3 = self.fc2(out2)
        out_middle = self.fc3(out3)
        out5 = self.fc4(out_middle)
        out6 = self.fc5(out5)
        output = self.fc6(out6)
        out7 = self.fc7(output)
        out8 = self.fc8(out7)

        out_label = self.fc9(out8)
        return x, out2, out3, out_middle, out5, out6, output, out7, out8, out_label


class Discriminator_attention_dualAE(nn.Module):
    def __init__(self, N1):
        super(Discriminator_attention_dualAE, self).__init__()
        self.N1 = N1
        self.softmax = nn.Softmax(dim=1)
        self.sigmoid = nn.Sigmoid()
        self.fc1 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc2 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc3 = nn.Linear(int(np.ceil(N1 / 25)), 1)

    def attention(self, a, b, input):
        middle_result = self.softmax(torch.mm(a, b.T)) / np.sqrt(a.size()[1])
        result = torch.mm(middle_result, input)
        return result

    def forward(self, x):
        out1 = self.fc1(x)
        # out1 = self.attention(out1, out1, out1)
        out2 = self.fc2(out1)
        out2 = self.attention(out2, out2, out2)
        out3 = self.fc3(out2)
        output = self.sigmoid(out3)
        return output


# ------ ------
# GAN + AE    (Generator :3coderAE)
# ------ ------
class Generator_3coder(nn.Module):

    def __init__(self, N1):
        super(Generator_3coder, self).__init__()
        self.N1 = N1
        self.fc1 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc2 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc3 = nn.Linear(int(np.ceil(N1 / 25)), 2)
        self.fc4 = nn.Linear(2, int(np.ceil(N1 / 25)))
        self.fc5 = nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5)))
        self.fc6 = nn.Linear(int(np.ceil(N1 / 5)), N1)
        self.fc7 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc8 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc9 = nn.Linear(int(np.ceil(N1 / 25)), 2)

    def forward(self, x):
        out2 = self.fc1(x)
        out3 = self.fc2(out2)
        out_middle = self.fc3(out3)
        out5 = self.fc4(out_middle)
        out6 = self.fc5(out5)
        output = self.fc6(out6)
        out7 = self.fc7(output)
        out8 = self.fc8(out7)
        out_label = self.fc9(out8)
        return x, out2, out3, out_middle, out5, out6, output, out7, out8, out_label


class Discriminator_3coder(nn.Module):
    def __init__(self, N1):
        super(Discriminator_3coder, self).__init__()
        self.N1 = N1
        self.encoder = nn.Sequential(
            nn.Linear(N1, int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25))),
            nn.Linear(int(np.ceil(N1 / 25)), 1),
            nn.Sigmoid()
        )

    def forward(self, x):
        output = self.encoder(x)
        return output


# ------ ------
# 3coderAE
# ------ ------

class three_coder_AE(nn.Module):

    def __init__(self, N1):
        super(three_coder_AE, self).__init__()
        self.N1 = N1
        self.fc1 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc2 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc3 = nn.Linear(int(np.ceil(N1 / 25)), 2)
        self.fc4 = nn.Linear(2, int(np.ceil(N1 / 25)))
        self.fc5 = nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5)))
        self.fc6 = nn.Linear(int(np.ceil(N1 / 5)), N1)
        self.fc7 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc8 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc9 = nn.Linear(int(np.ceil(N1 / 25)), 2)

    def forward(self, x):
        out2 = self.fc1(x)
        out3 = self.fc2(out2)
        out_middle = self.fc3(out3)
        out5 = self.fc4(out_middle)
        out6 = self.fc5(out5)
        output = self.fc6(out6)
        out7 = self.fc7(output)
        out8 = self.fc8(out7)
        out_label = self.fc9(out8)
        return x, out2, out3, out_middle, out5, out6, output, out7, out8, out_label


class unsupv_AE(nn.Module):

    def __init__(self, N1):
        super(unsupv_AE, self).__init__()
        self.N1 = N1
        self.fc1 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc2 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))

        self.fc3 = nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5)))
        self.fc4 = nn.Linear(int(np.ceil(N1 / 5)), N1)

    def forward(self, x):
        out2 = self.fc1(x)
        out_middle = self.fc2(out2)
        out4 = self.fc3(out_middle)
        output = self.fc4(out4)
        return x, out2, out_middle, out4, output


class supv_AE(nn.Module):

    def __init__(self, N1):
        super(supv_AE, self).__init__()
        self.N1 = N1
        self.encoder = nn.Sequential(
            nn.Linear(N1, int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        )
        self.middle = nn.Sequential(
            nn.Linear(int(np.ceil(N1 / 25)), 1),
            nn.Sigmoid()
        )
        self.decoder = nn.Sequential(
            nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), N1)
        )

    def forward(self, x):
        out1 = self.encoder(x)  # mid
        out2 = self.middle(out1)  # label
        out3 = self.decoder(out1)  # tail
        return out1, out2, out3


class Net2(nn.Module):
    def __init__(self, N1):
        super(Net2, self).__init__()
        self.encoder = nn.Sequential(
            nn.Linear(N1, int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25))),
            nn.Linear(int(np.ceil(N1 / 25)), 1),
            nn.Softmax()
        )

    def forward(self, x):
        out1 = self.encoder(x)
        return out1


# ============GAN+AE=============
class Generator(nn.Module):

    def __init__(self, N1):
        super(Generator, self).__init__()
        self.N1 = N1
        self.encoder = nn.Sequential(
            nn.Linear(N1, int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        )
        self.decoder = nn.Sequential(
            nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), N1)
        )

    def forward(self, x):
        out1 = self.encoder(x)  # mid
        out2 = self.decoder(out1)  # tail
        return out1, out2


class Discriminator(nn.Module):
    def __init__(self, Generator, N1):
        super(Discriminator, self).__init__()
        self.N1 = N1
        self.generator = Generator
        self.encoder = nn.Sequential(
            nn.Linear(N1, int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25))),
            nn.Linear(int(np.ceil(N1 / 25)), 1),
            nn.Sigmoid()
            # nn.Linear(int(np.ceil(N1 / 25)), 2),
            # nn.Softmax()
        )

    def forward(self, x):
        _, out1 = self.generator(x)
        out2 = self.encoder(out1)
        return out2


class siamese_AE(nn.Module):

    def __init__(self, N1):
        super(siamese_AE, self).__init__()
        self.N1 = N1
        self.encoder_1 = nn.Sequential(
            nn.Linear(N1, int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        )
        self.middle = nn.Linear(int(np.ceil(N1 / 25)) * 2, 2)
        self.decoder_1 = nn.Sequential(
            nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), N1)
        )
        self.encoder_2 = nn.Sequential(
            nn.Linear(N1, int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        )
        self.decoder_2 = nn.Sequential(
            nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), N1)
        )

    def forward(self, x1, x2):
        out_middle_1 = self.encoder_1(x1)  # 1mid
        out_middle_2 = self.encoder_2(x2)  # 2mid
        out3 = torch.cat((out_middle_1, out_middle_2), dim=1)  # 连接
        out4 = self.middle(out3)
        out_tail_1 = self.decoder_1(out_middle_1)  # 1mid
        out_tail_2 = self.decoder_2(out_middle_2)  # 2tail
        return out_middle_1, out_middle_2, out4, out_tail_1, out_tail_2


class attention_AE(nn.Module):

    def __init__(self, N1):
        super(attention_AE, self).__init__()
        self.N1 = N1
        self.softmax = nn.Softmax(dim=1)
        self.sigmoid = nn.Sigmoid()
        self.fc1 = nn.Linear(N1, int(np.ceil(N1 / 5)))
        self.fc2 = nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25)))
        self.fc3 = nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5)))
        self.fc4 = nn.Linear(int(np.ceil(N1 / 5)), N1)
        self.middle = nn.Linear(int(np.ceil(N1 / 25)), 1)

    def attention(self, a, b, input):
        middle_result = self.softmax(torch.mm(a, b.T)) / np.sqrt(a.size()[1])
        result = torch.mm(middle_result, input)
        return result

    def forward(self, x):
        out1 = self.fc1(x)
        out2 = self.fc2(out1)  # mid
        out3 = self.fc3(out2)

        output = self.fc4(out3)  # tile

        out4 = self.attention(x, output, x)
        out5 = self.attention(out1, out3, x)
        out6 = self.attention(out2, out2, x)
        out_sum = (out4 + out5 + out6) / 3
        out_label = self.sigmoid(self.middle(out2))
        return out2, out_sum, out_label


# ----------
# sample select
# ----------
'''

class sample_select_AE(nn.Module):

    def __init__(self, N1):
        super(sample_select_AE, self).__init__()
        self.N1 = N1  # 样本数目
        self.encoder = nn.Sequential(
            nn.Linear(N1, int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), int(np.ceil(N1 / 25))),
            nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 625)))
        )
        self.decoder = nn.Sequential(
            nn.Linear(int(np.ceil(N1 / 625)), int(np.ceil(N1 / 25))),
            nn.Linear(int(np.ceil(N1 / 25)), int(np.ceil(N1 / 5))),
            nn.Linear(int(np.ceil(N1 / 5)), N1)
        )

    def forward(self, x):
        out1 = self.encoder(x)  # mid
        out2 = self.decoder(out1)  # tail
        return out1, out2
'''