# !/usr/bin/env Python3
# -*- coding:utf-8 -*-
# Team:HiLab
# Author:Yu Tao
# Date:2022/7/9 17:34
# Tool:PyCharm
# Email:yuao_11260103@163.com
# Project:SCC
# ------ ------
# dual_AE as generator(with attention ).
# use wGAN+dual_AE+attention to feature construct.
# ------ ------

import torch
import random
import numpy as np
from scripts.model import setup_seed, wDiscriminator_attention_dualAE, wGenerator_attention_dualAE
import torch.nn.functional as F
from torch.utils.data import DataLoader
from param_class import Dataset_Param
from load_data import split_train_test
from torch import nn, optim
import matplotlib.pyplot as plt
from matplotlib.pyplot import subplots
from datetime import datetime

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
# Assuming that we are on a CUDA machine, this should print a CUDA device:
print(device)
# net.to(device)

seed = 19
setup_seed(seed)
torch.manual_seed(seed)
torch.cuda.manual_seed_all(seed)
np.random.seed(seed)
random.seed(seed)
torch.backends.cudnn.deterministic = True


class MyDataSet(torch.utils.data.Dataset):
    def __init__(self, data, label):
        self.data = torch.from_numpy(data).float()
        self.label = torch.from_numpy(label).float()  # torch.from_numpy(label)
        self.length = data.shape[0]

    def __getitem__(self, index):
        return self.data[index], self.label[index]

    def __len__(self):
        return self.length


def train_module(train_dataloader, gen, dis, optimizer_g, optimizer_d):
    print('----train moudle-----')
    num_batches = len(train_dataloader)
    batch_size = train_dataloader.batch_size
    train_d_loss = 0
    train_g_loss = 0
    gen.train()
    dis.train()
    for i, data in enumerate(train_dataloader, 0):
        for param in dis.parameters():
            param.data.clamp_(-0.01, 0.01)
        real_data, y = data
        real_data, y = real_data.to(device), y.to(device)

        # real_label = torch.autograd.Variable(torch.ones(batch_size)).to(device)
        # fake_label = torch.autograd.Variable(torch.zeros(batch_size)).to(device)

        _, _, _, _, _, _, fake_data, _, _, _ = gen(real_data)
        real_out = dis(real_data)
        real_out = real_out.view([batch_size])
        fake_out = dis(fake_data)
        fake_out = fake_out.view([batch_size])

        # real_loss = F.mse_loss(real_out, real_label)
        # fake_loss = F.mse_loss(fake_out, fake_label)
        # d_loss = (real_loss + fake_loss) / 2

        real_loss = -torch.mean(real_out)
        fake_loss = torch.mean(fake_out)
        d_loss = fake_loss + real_loss  # use  Wasserstein length

        train_d_loss += d_loss.item()

        optimizer_d.zero_grad()
        d_loss.backward()
        optimizer_d.step()

        _, _, _, _, _, _, output, _, _, out_label = gen(real_data)

        g_mse_loss = F.mse_loss(output, real_data)
        g_ce_loss = F.cross_entropy(out_label, y.long())
        g_loss = (g_mse_loss + g_ce_loss) / 2

        train_g_loss += g_loss.item()

        optimizer_g.zero_grad()
        g_loss.backward()
        optimizer_g.step()

    return train_d_loss / num_batches, train_g_loss / num_batches


def test_module(test_dataloader, gen, dis):
    print('----test moudle-----')
    num_batches = len(test_dataloader)
    test_g_loss = 0
    test_d_loss = 0
    batch_size = test_dataloader.batch_size
    with torch.no_grad():
        for i, data in enumerate(test_dataloader, 0):
            real_data, y = data
            real_data, y = real_data.to(device), y.to(device)
            # real_label = torch.autograd.Variable(torch.ones(batch_size)).to(device)
            # fake_label = torch.autograd.Variable(torch.zeros(batch_size)).to(device)
            _, _, _, _, _, _, fake_data, _, _, _ = gen(real_data)
            real_out = dis(real_data)
            real_out = real_out.view([batch_size])
            fake_out = dis(fake_data)
            fake_out = fake_out.view([batch_size])
            # real_loss = F.mse_loss(real_out, real_label)
            # fake_loss = F.mse_loss(fake_out, fake_label)
            # d_loss = (real_loss + fake_loss) / 2

            real_loss = -torch.mean(real_out)
            fake_loss = torch.mean(fake_out)
            d_loss = fake_loss + real_loss  # use  Wasserstein length

            test_d_loss += d_loss.item()

            _, _, _, _, _, _, output, _, _, out_label = gen(real_data)
            g_mse_loss = F.mse_loss(output, real_data)
            g_ce_loss = F.cross_entropy(out_label, y.long())
            g_loss = (g_mse_loss + g_ce_loss) / 2

            test_g_loss += g_loss.item()

    return test_d_loss / num_batches, test_g_loss / num_batches


def wgan_attention_dualAE_FC(X, y, Dataset_Param):
    print('---start gan_3coderAE_FC---')
    X_train, X_test, y_train, y_test = split_train_test(X, y)
    data_dict = {'X_train': X_train, 'X_test': X_test, 'y_train': y_train, 'y_test': y_test}

    print("data loading...")
    train_data_set = MyDataSet(data_dict['X_train'], data_dict['y_train'])
    train_dataloader = DataLoader(train_data_set, batch_size=Dataset_Param.batch_size, shuffle=True, drop_last=True)
    test_data_set = MyDataSet(data_dict['X_test'], data_dict['y_test'])
    test_dataloader = DataLoader(test_data_set, batch_size=Dataset_Param.batch_size, shuffle=True, drop_last=True)

    print('model constructing....')
    N1 = data_dict['X_train'].shape[1]
    gen = wGenerator_attention_dualAE(N1)
    dis = wDiscriminator_attention_dualAE(N1)
    gen.to(device)
    dis.to(device)
    # use RMSprop optimizer
    optimizer_g = torch.optim.RMSprop(gen.parameters(), lr=Dataset_Param.lr)
    optimizer_d = torch.optim.RMSprop(dis.parameters(), lr=Dataset_Param.lr)
    g_trains = []
    d_trains = []
    g_tests = []
    d_tests = []

    for i in range(Dataset_Param.epoch):  # epoch
        print(i)
        d_train, g_train = train_module(train_dataloader, gen, dis, optimizer_g, optimizer_d)
        d_test, g_test = test_module(test_dataloader, gen, dis)
        g_trains.append(g_train)
        d_trains.append(d_train)
        g_tests.append(g_test)
        d_tests.append(d_test)

    print("OK")
    # ---观察损失曲线---
    print("start drawing...")
    fig, axes = subplots(1, 1)
    axes.plot(g_trains, label='g_trains')
    axes.plot(d_trains, label='d_trains')
    axes.plot(g_tests, label='g_tests')
    axes.plot(d_tests, label='d_tests')
    axes.set_xlabel('epoch', fontsize=13)
    plt.title(Dataset_Param.name + '_' + str(Dataset_Param.batch_size) + '_' + str(
        Dataset_Param.lr) + '-trainset-wGAN_attention_dualAE_FC')
    plt.legend(fontsize=12)
    plt.savefig(
        'fig/wGAN_attention_dualAE/' + Dataset_Param.name + '_' + str(Dataset_Param.batch_size) + '_' + str(
            Dataset_Param.lr) + '_' + str(
            Dataset_Param.epoch) + '.png')
    return gen
